# coding: utf8

# Copyright 2014-2015 Vincent Jacques <vincent@vincent-jacques.net>

from ...batch_get_item import BatchGetItemLocalIntegTests
from ...batch_write_item import BatchWriteItemLocalIntegTests
from ...create_table import CreateTableLocalIntegTests, CreateTableErrorLocalIntegTests
from ...delete_item import DeleteItemLocalIntegTests
from ...delete_table import DeleteTableLocalIntegTests
from ...describe_table import DescribeTableLocalIntegTests
from ...get_item import GetItemLocalIntegTests
from ...list_tables import ListTablesLocalIntegTests
from ...put_item import PutItemLocalIntegTests
from ...query import QueryLocalIntegTests
from ...scan import ScanLocalIntegTests
from ...update_item import UpdateItemLocalIntegTests
from ...update_table import UpdateTableLocalIntegTests
