# coding: utf8

# Copyright 2014-2015 Vincent Jacques <vincent@vincent-jacques.net>

from ...batch_delete_item import BatchDeleteItemLocalIntegTests
from ...batch_get_item_iterator import BatchGetItemIteratorLocalIntegTests
from ...batch_put_item import BatchPutItemLocalIntegTests
from ...list_tables_iterator import ListTablesIteratorLocalIntegTests
from ...query_iterator import QueryIteratorLocalIntegTests
from ...scan_iterator import ScanIteratorLocalIntegTests
from ...wait_for_table_activation import WaitForTableActivationLocalIntegTests
from ...wait_for_table_deletion import WaitForTableDeletionLocalIntegTests
