# coding: utf8

# Copyright 2014-2015 Vincent Jacques <vincent@vincent-jacques.net>

from ...batch_get_item import BatchGetItemConnectedIntegTests
from ...batch_write_item import BatchWriteItemConnectedIntegTests
from ...delete_item import DeleteItemConnectedIntegTests
from ...get_item import GetItemConnectedIntegTests
from ...put_item import PutItemConnectedIntegTests
from ...query import QueryConnectedIntegTests
from ...scan import ScanConnectedIntegTests
from ...update_item import UpdateItemConnectedIntegTests
