# coding: utf8

# Copyright 2014-2015 Vincent Jacques <vincent@vincent-jacques.net>

from ..conversion import ConversionUnitTests
from ..expressions import ConditionExpressionUnitTests

from ..batch_get_item import BatchGetItemUnitTests
from ..batch_write_item import BatchWriteItemUnitTests
from ..create_table import CreateTableUnitTests
from ..delete_item import DeleteItemUnitTests
from ..delete_table import DeleteTableUnitTests
from ..describe_table import DescribeTableUnitTests
from ..get_item import GetItemUnitTests
from ..list_tables import ListTablesUnitTests
from ..put_item import PutItemUnitTests
from ..query import QueryUnitTests
from ..scan import ScanUnitTests
from ..update_item import UpdateItemUnitTests
from ..update_table import UpdateTableUnitTests
