# coding: utf8

# Copyright 2014-2015 Vincent Jacques <vincent@vincent-jacques.net>

import LowVoltage.testing as _tst
from LowVoltage.actions.tests.unit import *
from LowVoltage.compounds.tests.unit import *
from LowVoltage.connection.tests.unit import *


if __name__ == "__main__":  # pragma no branch (Test code)
    _tst.main()  # pragma no cover (Test code)
