# coding: utf8

# Copyright 2014-2015 Vincent Jacques <vincent@vincent-jacques.net>

from ..connection import ConnectionUnitTests, SignerUnitTests, ResponderUnitTests
from ..credentials import Ec2RoleCredentialsUnitTests
from ..retry_policies import ExponentialBackoffRetryPolicyUnitTests
